
#include "platform.h"
#include "videosystem.h"
#include "core.h"
#include "algorithms.h"
#include <math.h>

// out must be passed uninitialized !
// sigma -> factor in calculation
//
// return codes:
//		 0		success
//		-1		out was not passed uninitialized
//		-2		signal_and_background must at least contain 2 frames
//		-3		background must at least contain 2 frames
//		-4		signal_and_bck, background must contain the same number of images!
//		-5		bitdepth, effective bitdepth, width or heigt of background and signal_and_background were
//				not set to equal values (important)
//		-6		memory allocation error, no data was changed (in out)	
//		-7		data error (unsupported bpp for algorithm)
//

INT32 calculate_significance_map( const struct VIDINFHDR *signal_and_background, const struct VIDINFHDR *background, struct VIDINFHDR *out, DOUBLE sigma)
{
	if (out->type != TYPE_NONE_EMPTY) return (-1);

	if (signal_and_background->numframes < 2) return -2;
	if (background->numframes < 2) return -3;

	if (signal_and_background->numframes != background->numframes) return -4;


	if ( (signal_and_background->width != background->width) ||
		 (signal_and_background->height != background->height) ||
		 (signal_and_background->bpp != background->bpp) ||
		 (signal_and_background->bpp_effective != background->bpp_effective) ) return -5;

	if ((signal_and_background->bpp < 1) || (signal_and_background->bpp >32)) return -7;

	// now we can start

	{
		UINT32 numframes = signal_and_background->numframes;
		UINT32 bpp = signal_and_background->bpp;
		UINT32 i,j;
		UINT32 size_pixels = signal_and_background->width*signal_and_background->height;

		DOUBLE *sig_xmeanbuffer;
		DOUBLE *back_xmeanbuffer;
		DOUBLE  numimg_minus_1 = (numframes-1)*1.0;
		DOUBLE *rms_noise_buf;
		VOID   *resulting_buf;

		resulting_buf = (VOID *) malloc( size_pixels*bpp/8 );
		if (resulting_buf == NULL)
		{
			return -6;
		}

		memset(resulting_buf, 0, size_pixels*bpp/8);

		sig_xmeanbuffer = (DOUBLE *) malloc( size_pixels*sizeof(DOUBLE) );
		if (sig_xmeanbuffer == NULL) 
		{
			free( (void *) resulting_buf );
			return -6;
		}

		back_xmeanbuffer = (DOUBLE *) malloc( size_pixels*sizeof(DOUBLE) );
		if (back_xmeanbuffer == NULL) 
		{
			free( (void *) sig_xmeanbuffer );
			free( (void *) resulting_buf );
		
			return -6;
		}

		rms_noise_buf = (DOUBLE *) malloc( size_pixels*sizeof(DOUBLE) );
		if (rms_noise_buf == NULL) 
		{
			free( (void *) back_xmeanbuffer );
			free( (void *) sig_xmeanbuffer );
			free( (void *) resulting_buf );				

			return -6;
		}

		if (bpp <= 8)
		{
			// calculate xmean of signal+noise and of noise
			UCHAR *sig_bitmap = signal_and_background->framebits[0];
			UCHAR *back_bitmap = background->framebits[0];

			for (j=0;j<size_pixels;j++)
			{
				sig_xmeanbuffer[j] = (DOUBLE) (sig_bitmap[j]);
			}


			for (i=1;i<numframes;i++)
			{
				sig_bitmap = signal_and_background->framebits[i];
				
				for (j=0;j<size_pixels;j++)
				{
					sig_xmeanbuffer[j] += (DOUBLE) (sig_bitmap[j]);
				}
			}

			for (j=0;j<size_pixels;j++)
			{
				sig_xmeanbuffer[j] /= (DOUBLE) (numframes);
			}


			for (j=0;j<size_pixels;j++)
			{
				back_xmeanbuffer[j] = (DOUBLE) (back_bitmap[j]);
			}

			for (i=1;i<numframes;i++)
			{
				back_bitmap = background->framebits[i];
				
				for (j=0;j<size_pixels;j++)
				{
					back_xmeanbuffer[j] += (DOUBLE) (back_bitmap[j]);
				}
			}

			for (j=0;j<size_pixels;j++)
			{
				back_xmeanbuffer[j] /= (DOUBLE) (numframes);
			}

			// calculate RMS of noise

			for (j=0;j<size_pixels;j++)
			{
				DOUBLE bck_xmean = back_xmeanbuffer[j];
				DOUBLE result;

				result = 0.0;
				for (i=0;i<numframes;i++)
				{
					UCHAR *back_bitmap = background->framebits[i];
					DOUBLE result1 = bck_xmean - ((DOUBLE) back_bitmap[j]);

					result += ((result1*result1));
				}

				rms_noise_buf[j] = sigma*sqrt(result/numimg_minus_1);
			}

			{
				UCHAR *lresbuf = (UCHAR *) resulting_buf;

				// create resulting data set 
				for (j=0;j<size_pixels;j++)
				{
					if ((sig_xmeanbuffer[j]-back_xmeanbuffer[j])<=(rms_noise_buf[j])) lresbuf[j] = 1;
				}
			}

		}
		else if (bpp <= 16)
		{
			// calculate xmean of signal+noise and of noise
			UINT16 *sig_bitmap = signal_and_background->framebits[0];
			UINT16 *back_bitmap = background->framebits[0];

			for (j=0;j<size_pixels;j++)
			{
				sig_xmeanbuffer[j] = (DOUBLE) (sig_bitmap[j]);
			}


			for (i=1;i<numframes;i++)
			{
				sig_bitmap = signal_and_background->framebits[i];
				
				for (j=0;j<size_pixels;j++)
				{
					sig_xmeanbuffer[j] += (DOUBLE) (sig_bitmap[j]);
				}
			}

			for (j=0;j<size_pixels;j++)
			{
				sig_xmeanbuffer[j] /= (DOUBLE) (numframes);
			}


			for (j=0;j<size_pixels;j++)
			{
				back_xmeanbuffer[j] = (DOUBLE) (back_bitmap[j]);
			}

			for (i=1;i<numframes;i++)
			{
				back_bitmap = background->framebits[i];
				
				for (j=0;j<size_pixels;j++)
				{
					back_xmeanbuffer[j] += (DOUBLE) (back_bitmap[j]);
				}
			}

			for (j=0;j<size_pixels;j++)
			{
				back_xmeanbuffer[j] /= (DOUBLE) (numframes);
			}

			// calculate RMS of noise

			for (j=0;j<size_pixels;j++)
			{
				DOUBLE bck_xmean = back_xmeanbuffer[j];
				DOUBLE result;

				result = 0.0;
				for (i=0;i<numframes;i++)
				{
					UINT16 *back_bitmap = background->framebits[i];
					DOUBLE result1 = bck_xmean - ((DOUBLE) back_bitmap[j]);

					result += ((result1*result1));
				}

				rms_noise_buf[j] = sigma*sqrt(result/numimg_minus_1);
			}

			{
				UINT16 *lresbuf = (UINT16 *) resulting_buf;

				// create resulting data set 
				for (j=0;j<size_pixels;j++)
				{
					if ((sig_xmeanbuffer[j]-back_xmeanbuffer[j])<=(rms_noise_buf[j])) lresbuf[j] = 1;
				}
			}

		}
		else if (bpp <= 32)
		{
			// calculate xmean of signal+noise and of noise
			UINT32 *sig_bitmap = signal_and_background->framebits[0];
			UINT32 *back_bitmap = background->framebits[0];

			for (j=0;j<size_pixels;j++)
			{
				sig_xmeanbuffer[j] = (DOUBLE) (sig_bitmap[j]);
			}


			for (i=1;i<numframes;i++)
			{
				sig_bitmap = signal_and_background->framebits[i];
				
				for (j=0;j<size_pixels;j++)
				{
					sig_xmeanbuffer[j] += (DOUBLE) (sig_bitmap[j]);
				}
			}

			for (j=0;j<size_pixels;j++)
			{
				sig_xmeanbuffer[j] /= (DOUBLE) (numframes);
			}


			for (j=0;j<size_pixels;j++)
			{
				back_xmeanbuffer[j] = (DOUBLE) (back_bitmap[j]);
			}

			for (i=1;i<numframes;i++)
			{
				back_bitmap = background->framebits[i];
				
				for (j=0;j<size_pixels;j++)
				{
					back_xmeanbuffer[j] += (DOUBLE) (back_bitmap[j]);
				}
			}

			for (j=0;j<size_pixels;j++)
			{
				back_xmeanbuffer[j] /= (DOUBLE) (numframes);
			}

			// calculate RMS of noise

			for (j=0;j<size_pixels;j++)
			{
				DOUBLE bck_xmean = back_xmeanbuffer[j];
				DOUBLE result;

				result = 0.0;
				for (i=0;i<numframes;i++)
				{
					UINT32 *back_bitmap = background->framebits[i];
					DOUBLE result1 = bck_xmean - ((DOUBLE) back_bitmap[j]);

					result += ((result1*result1));
				}

				rms_noise_buf[j] = sigma*sqrt(result/numimg_minus_1);
			}

			{
				UINT32 *lresbuf = (UINT32 *) resulting_buf;

				// create resulting data set 
				for (j=0;j<size_pixels;j++)
				{
					if ((sig_xmeanbuffer[j]-back_xmeanbuffer[j])<=(rms_noise_buf[j])) lresbuf[j] = 1;
				}
			}
		}
		else return -7; // no proper bpp setting

		out->framescales = (DOUBLE *) malloc( 1*sizeof(DOUBLE) );
		if (out->framescales == NULL)
		{
			free( (void *) rms_noise_buf );
			free( (void *) back_xmeanbuffer );
			free( (void *) sig_xmeanbuffer );
			free( resulting_buf );

			return -6;
		}

		out->framebits = (VOID **) malloc( sizeof( VOID **)*1 );
		if (out->framebits == NULL)
		{
			free( (void *) out->framescales );
			free( (void *) rms_noise_buf );
			free( (void *) back_xmeanbuffer );
			free( (void *) sig_xmeanbuffer );
			free( resulting_buf );

			return -6;
		}

		out->framebits[0] = (void *) resulting_buf;
		out->framescales[0] = signal_and_background->scale;

		out->bpp = signal_and_background->bpp;
		out->bpp_effective = signal_and_background->bpp_effective;
		out->scale = signal_and_background->scale;
		out->type = TYPE_SIGNIFICANCE;
		out->width = signal_and_background->width;
		out->height = signal_and_background->height;
		out->numframes = 1;

		free( (void *) rms_noise_buf );
		free( (void *) back_xmeanbuffer );
		free( (void *) sig_xmeanbuffer );

	}


	return 0;
}


// macro that will be used to keep functions clean and tidy when
// reacting on error conditions
#define ERR(a)  \
		 { errcode = a; \
		goto err;}  \


// This function is used to downscale a vinf header containing multiple images to 
// a vinf header containing a combined, single image. The single image is created 
// out of the multiple images by applying a envelope function.
// 
// This function is helpful e.g. when converting BKC to BKG file.
//
//	Parameters:
//		[in/out]	vinf		VIDINFHDR containing multiple images [in]
//								VIDINFHDR containing the downscaled single image result [out]
//
//	Return values:
//		 0		done, finished successfully
//		-5		bpp not supported
//		-6		memory allocation error
//
//	Note: In any error case the vinf is not touched and the old data remains inside 
//		  (without corruption) !
//
INT32 downscale_envelope_function( struct VIDINFHDR *vinf )
{
	UINT32 *buffer32 = (UINT32 *) malloc( vinf->width*vinf->height*sizeof(UINT32) );
	INT32 num;
	UINT32 x;
	INT32 errcode=0;
	DOUBLE *local_framescales=NULL;
	VOIDP *local_framebits=NULL;
		
	if (buffer32 == NULL) ERR(-6);

	for (x=0;x<(vinf->width*vinf->height);x++)
	{
		buffer32[x] = 0;
	}
			
	for (num=0;num<vinf->numframes;num++)
	{
		if (vinf->bpp <= 8)
		{
			UCHAR *imgbuf = (UCHAR *) vinf->framebits[num];
		
			for (x=0;x<(vinf->width*vinf->height);x++)
			{
				UINT32 d = (UINT32) *imgbuf++;
				if (buffer32[x] < d) buffer32[x] = d;
			}
		}
		else if (vinf->bpp <= 16)
		{
			UINT16 *imgbuf = (UINT16 *) vinf->framebits[num];
		
			for (x=0;x<(vinf->width*vinf->height);x++)
			{
				UINT32 d = (UINT32) *imgbuf++;
				if (buffer32[x] < d) buffer32[x] = d;
			}
		}
		else if (vinf->bpp <= 32)
		{
			UINT32 *imgbuf = (UINT32 *) vinf->framebits[num];
		
			for (x=0;x<(vinf->width*vinf->height);x++)
			{
				UINT32 d = *imgbuf++;
				if (buffer32[x] < d) buffer32[x] = d;
			}
		}
		else ERR( -5 );
	}
	
	// drop back buffer, adjust VIDINFHDR
	
	local_framescales = (DOUBLE *) malloc( sizeof(DOUBLE)*1 );
	if (local_framescales == NULL) ERR(-6);
		
	local_framebits = (VOIDP *) malloc( sizeof(VOIDP *)*1 );
	if (local_framebits == NULL) ERR(-6);
	
	local_framebits[0] = (VOIDP) malloc( (vinf->width*vinf->height*vinf->bpp)/8 );
	if (local_framebits[0] == NULL) ERR(-6);


	// free old memory buffers	
	
	if (vinf->framescales != NULL) free( vinf->framescales );
	
	for (num=0;num<vinf->numframes;num++)
	{
		if (vinf->framebits[num] != NULL) free( (VOIDP) vinf->framebits[num] );
	}
	if (vinf->framebits != NULL) free( (VOIDP) vinf->framebits );
	
	// adjjust vinf header
	vinf->numframes = 1;	
	vinf->framescales = local_framescales;
	vinf->framebits = local_framebits;
	vinf->framescales[0] = vinf->scale;
	
	// update data	
		
	if (vinf->bpp <= 8)
	{
		UCHAR *imgbuf = (UCHAR *) vinf->framebits[0];
	
		for (x=0;x<(vinf->width*vinf->height);x++)
		{
			*imgbuf++ = (UCHAR) buffer32[x];
		}
	}
	else if (vinf->bpp <= 16)
	{
		UINT16 *imgbuf = (UINT16 *) vinf->framebits[0];
	
		for (x=0;x<(vinf->width*vinf->height);x++)
		{
			*imgbuf++ = (UINT16) buffer32[x];
		}
	}
	else if (vinf->bpp <= 32)
	{
		UINT32 *imgbuf = (UINT32 *) vinf->framebits[0];
	
		for (x=0;x<(vinf->width*vinf->height);x++)
		{
			*imgbuf++ = buffer32[x];				
		}
	}
	
	
err:
	if (errcode != 0)
	{
  		if (local_framescales != NULL) free( (VOIDP) local_framescales );
  		if (local_framebits != NULL)
  		{
  			if (local_framebits[0] != NULL) free( (VOIDP) local_framebits[0] );
  			free( (VOIDP) local_framebits );
  		}
	}
	
	free( (VOIDP) buffer32 );

	return errcode;
}

// This function is used to downscale a vinf header containing multiple images to 
// a vinf header containing a combined, single image. The single image is created 
// out of the multiple images by applying an 'averaging' function.
// 
// This function is helpful e.g. when converting BKC to BKG file.
//
//	Parameters:
//		[in/out]	vinf		VIDINFHDR containing multiple images [in]
//								VIDINFHDR containing the downscaled single image result [out]
//
//	Return values:
//		 0		done, finished successfully
//		-5		bpp not supported
//		-6		memory allocation error
//
//	Note: In any error case the vinf is not touched and the old data remains inside 
//		  (without corruption) !
//
INT32 downscale_average_function( struct VIDINFHDR *vinf )
{
	DOUBLE *bufferD = (DOUBLE *) malloc( vinf->width*vinf->height*sizeof(DOUBLE) );
	INT32 num;
	UINT32 x;
	INT32 errcode=0;
	DOUBLE *local_framescales=NULL;
	VOIDP *local_framebits=NULL;
	DOUBLE old_num_frames = (DOUBLE) vinf->numframes;
	
	if (bufferD == NULL) ERR(-6);

	for (x=0;x<(vinf->width*vinf->height);x++)
	{
		bufferD[x] = 0.0;
	}
			
	for (num=0;num<vinf->numframes;num++)
	{
		if (vinf->bpp <= 8)
		{
			UCHAR *imgbuf = (UCHAR *) vinf->framebits[num];
		
			for (x=0;x<(vinf->width*vinf->height);x++)
			{
				bufferD[x] += (DOUBLE) *imgbuf++;
			}
		}
		else if (vinf->bpp <= 16)
		{
			UINT16 *imgbuf = (UINT16 *) vinf->framebits[num];
		
			for (x=0;x<(vinf->width*vinf->height);x++)
			{
				bufferD[x] += (DOUBLE) *imgbuf++;				
			}
		}
		else if (vinf->bpp <= 32)
		{
			UINT32 *imgbuf = (UINT32 *) vinf->framebits[num];
		
			for (x=0;x<(vinf->width*vinf->height);x++)
			{
				bufferD[x] += (DOUBLE) *imgbuf++;				
			}
		}
		else ERR( -5 );
	}
	
	// drop back buffer, adjust VIDINFHDR
	
	local_framescales = (DOUBLE *) malloc( sizeof(DOUBLE)*1 );
	if (local_framescales == NULL) ERR(-6);
		
	local_framebits = (VOIDP *) malloc( sizeof(VOIDP *)*1 );
	if (local_framebits == NULL) ERR(-6);
	
	local_framebits[0] = (VOIDP) malloc( (vinf->width*vinf->height*vinf->bpp)/8 );
	if (local_framebits[0] == NULL) ERR(-6);


	// free old memory buffers	
	
	if (vinf->framescales != NULL) free( vinf->framescales );
	
	for (num=0;num<vinf->numframes;num++)
	{
		if (vinf->framebits[num] != NULL) free( (VOIDP) vinf->framebits[num] );
	}
	if (vinf->framebits != NULL) free( (VOIDP) vinf->framebits );
	
	// adjjust vinf header
	vinf->numframes = 1;	
	vinf->framescales = local_framescales;
	vinf->framebits = local_framebits;
	vinf->framescales[0] = vinf->scale;
	
	// update data	
		
	if (vinf->bpp <= 8)
	{
		UCHAR *imgbuf = (UCHAR *) vinf->framebits[0];
		
		for (x=0;x<(vinf->width*vinf->height);x++)
		{
			*imgbuf++ = (UCHAR) (bufferD[x] / old_num_frames);
		}
	}
	else if (vinf->bpp <= 16)
	{
		UINT16 *imgbuf = (UINT16 *) vinf->framebits[0];
		
		for (x=0;x<(vinf->width*vinf->height);x++)
		{
			*imgbuf++ = (UINT16) (bufferD[x] / old_num_frames);
		}
	}
	else if (vinf->bpp <= 32)
	{
		UINT32 *imgbuf = (UINT32 *) vinf->framebits[0];
		
		for (x=0;x<(vinf->width*vinf->height);x++)
		{
			*imgbuf++ = (UINT32) (bufferD[x] / old_num_frames);
		}
	}
	
	
err:
	if (errcode != 0)
	{
  		if (local_framescales != NULL) free( (VOIDP) local_framescales );
  		if (local_framebits != NULL)
  		{
  			if (local_framebits[0] != NULL) free( (VOIDP) local_framebits[0] );
  			free( (VOIDP) local_framebits );
  		}
	}
	
	free( (VOIDP) bufferD );

	return errcode;
}


// video_type -> FILE_IMM, FILE_IMC, FILE_BKG, FILE_BKC

// downscale-function: FCN_AVERAGE, FCN_ENVELOPE to create an enveloped or averaged background IF
// the resulting file format will be FILE_BKG (1 image only!)
// FCN_NONE otherwise, if any other type is requested, downscale_function must be FCN_NONE !!

//
// This function is used to conveniently convert between different formats inside a VIDINFHDR. Most conversions
// are lossless, however, conversion of a VIDINFHDR with multiple images to FILE_BKG will be lossy because a
// downscale function is required and will be used.
//
//	Parameters:
//		[in/out]	vinf				VIDINFHDR containing multiple images [in]
//										VIDINFHDR containing the downscaled single image result [out]
//		[in]		newtype				new type inside VIDINFHDR (one of FILE_IMM, FILE_IMC, FILE_BKG, FILE_BKC)
//		[in]		downscale_function	if the resulting format will be FILE_BKG, a downscale function must be passed
//										in order to recalculate the contents. either FCN_AVERAGE or FCN_ENVELOPE might
//										be passed. In any other case one must pass the FCN_NONE statement.
//	Return values:
//
//		 0		okay, everything fine
//		-1		inappropriate 'type' inside vinf structre
//		-2		inappropriate 'type' parameter passed
//		-3		downscale function selected which does not fit to passed structure and resulting type
//		-4		convert to bkg, downscale function was NOT specified
//		-5		unsupported bits per pixel settings inside vinf (using downscaling)
//		-6		memory allocation error (out of memory, using downscale to bkg)
//
// Remarks:
//	In any error case the vinf is not touched and the old data remains inside (without corruption) !
//	FCN_AVERAGE		averaging each frame pixel over the number of frames (average noise rules)
//  FCN_ENVELOPE	the maximum value of each pixel for all frames is obtained and stored as 
//					bitmap (max noise rules)
//
VS_EXPORT INT32 convert_format( struct VIDINFHDR *vinf, UINT32 newtype, UINT32 downscale_function)
{
	UINT32 ret;
	
	if ((vinf->type != TYPE_IMM) && (vinf->type != TYPE_IMC) && (vinf->type != TYPE_BKC) && (vinf->type != TYPE_BKG))
	{
		return( -1 );
	}

	switch( newtype )
	{
		case TYPE_IMM:
				if (downscale_function != FCN_NONE) return( -3 );
				break;
				
		case TYPE_IMC:
				if (downscale_function != FCN_NONE) return( -3 );
				break;

		case TYPE_BKC:
				if (downscale_function != FCN_NONE) return( -3 );
				break;

		case TYPE_BKG:
				if (vinf->numframes > 1)
				{
					if (downscale_function == FCN_ENVELOPE)
					{
						ret = downscale_envelope_function( vinf );
						if (ret != 0) return( ret );
					}
					else if (downscale_function == FCN_AVERAGE) 
					{
						ret = downscale_average_function( vinf );
						if (ret != 0) return( ret );
					}
					else if (downscale_function == FCN_NONE) return( -4 );
					else return( -3 );
				}
				break;

		default:
				return( -2 );
				break;
	}

	vinf->type = newtype;
	return 0; // all fine
} // convert_format()


// return value(s)
//	0 - okay, everything fine
// -1 - bad source data in vinout, data in vinout was NOT changed
//
VS_EXPORT INT32 xrayfilter( struct VIDINFHDR *vinout)
{
	// muss einfach nur bilder enthalten

	INT32 up = 0;
	INT32 left = 0;
	INT32 ImageWidth = vinout->width;
	INT32 ImageHeight = vinout->height;
	INT32 down = ImageHeight - 1;
	INT32 right = ImageWidth - 1;
	INT32 curimg;
	INT32 pos;

	if (vinout->type == TYPE_NONE_EMPTY) return -1;
	if (((down-up)<=2) || ((right-left)<=2)) return -1; // 3x3 matrix!

	for (curimg=0;curimg<vinout->numframes;curimg++)
	{
		if (vinout->bpp == 8)
		{
			INT32 j,i;
			UCHAR *data8 = (UCHAR *) vinout->framebits[curimg];
		
			for(j=up+1;j<=down-1;j++)
			{
				pos = j*ImageWidth+left+1; 
	 
				for(i=left+1;i<=right-1;i++)
				{
					INT32 equalize=0;
					const INT32 datapos = data8[pos];

					if (datapos)
					{
						const INT32 datapos2 = (datapos * 2) / 3;
						//lr = -1;
						//ub = -1;
						//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
						equalize+=((data8[pos-ImageWidth-1])<datapos2);

						//lr = -1;
						//ub = 0;
						//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
						equalize+=((data8[pos-1])<datapos2);

						//lr = -1;
						//ub = 1;
						//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
						equalize+=((data8[pos+ImageWidth-1])<datapos2);

						//lr=0;
						//ub = -1;
						//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
						equalize+=((data8[pos-ImageWidth])<datapos2);

						//lr = 0;
						//ub = 1;
						//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
						equalize+=((data8[pos+ImageWidth])<datapos2);

						//lr = 1;
						//ub = -1;
						//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
						equalize+=((data8[pos-ImageWidth+1])<datapos2);

						//lr = 1;
						//ub = 0;
						//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
						equalize+=((data8[pos+1])<datapos2);

						//lr = 1;
						//ub = 1;
						//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
						equalize+=((data8[pos+ImageWidth+1])<datapos2);


						if(equalize>3) // equalize
						{
							UINT32 sum=0;

  							//lr = -1;
							//ub = -1;
							//sum+=data[(i+lr)+(j+ub)*ImageWidth];
							sum+=data8[pos-ImageWidth-1];

							//lr = -1;
							//ub = 0;
							//sum+=data[(i+lr)+(j+ub)*ImageWidth];
							sum+=data8[pos-1];

							//lr = -1;
							//ub = 1;
							//sum+=data[(i+lr)+(j+ub)*ImageWidth];
							sum+=data8[pos+ImageWidth-1];

							//lr=0;
							//ub = -1;
							//sum+=data[(i+lr)+(j+ub)*ImageWidth];
							sum+=data8[pos-ImageWidth];

							//lr = 0;
							//ub = 1;
							//sum+=data[(i+lr)+(j+ub)*ImageWidth];
							sum+=data8[pos+ImageWidth];

							//lr = 1;
							//ub = -1;
							//sum+=data[(i+lr)+(j+ub)*ImageWidth];
							sum+=data8[pos-ImageWidth+1];

							//lr = 1;
							//ub = 0;
							//sum+=data[(i+lr)+(j+ub)*ImageWidth];
							sum+=data8[pos+1];

							//lr = 1;
							//ub = 1;
							//sum+=data[(i+lr)+(j+ub)*ImageWidth];
							sum+=data8[pos+ImageWidth+1];
				
							//sum+=data[(i+lr)+(j+ub)*ImageWidth];
				
							sum >>= 3;
				
							data8[pos]=(UCHAR) sum;
						} // equalize
					}
					pos++;
				}
			}
		}
		else if (vinout->bpp == 16)
		{
			INT32 j,i;
			UINT16 *data16 = (UINT16 *) vinout->framebits[curimg];
			for(j=up+1;j<=down-1;j++)
			{
				pos = j*ImageWidth+left+1;
	 
				for(i=left+1;i<=right-1;i++)
				{
					INT32 equalize=0;
					const INT32 datapos = data16[pos];

					if (datapos)
					{
						const UINT32 datapos2 = (datapos * 2) / 3;
						//lr = -1;
						//ub = -1;
						//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
						equalize+=((data16[pos-ImageWidth-1])<datapos2);

						//lr = -1;
						//ub = 0;
						//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
						equalize+=((data16[pos-1])<datapos2);

						//lr = -1;
						//ub = 1;
						//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
						equalize+=((data16[pos+ImageWidth-1])<datapos2);

						//lr=0;
						//ub = -1;
						//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
						equalize+=((data16[pos-ImageWidth])<datapos2);

						//lr = 0;
						//ub = 1;
						//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
						equalize+=((data16[pos+ImageWidth])<datapos2);

						//lr = 1;
						//ub = -1;
						//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
						equalize+=((data16[pos-ImageWidth+1])<datapos2);

						//lr = 1;
						//ub = 0;
						//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
						equalize+=((data16[pos+1])<datapos2);

						//lr = 1;
						//ub = 1;
						//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
						equalize+=((data16[pos+ImageWidth+1])<datapos2);


						if (equalize>3) // equalize
						{
							UINT32 sum=0;

  							//lr = -1;
							//ub = -1;
							//sum+=data[(i+lr)+(j+ub)*ImageWidth];
							sum+=data16[pos-ImageWidth-1];

							//lr = -1;
							//ub = 0;
							//sum+=data[(i+lr)+(j+ub)*ImageWidth];
							sum+=data16[pos-1];

							//lr = -1;
							//ub = 1;
							//sum+=data[(i+lr)+(j+ub)*ImageWidth];
							sum+=data16[pos+ImageWidth-1];

							//lr=0;
							//ub = -1;
							//sum+=data[(i+lr)+(j+ub)*ImageWidth];
							sum+=data16[pos-ImageWidth];

							//lr = 0;
							//ub = 1;
							//sum+=data[(i+lr)+(j+ub)*ImageWidth];
							sum+=data16[pos+ImageWidth];

							//lr = 1;
							//ub = -1;
							//sum+=data[(i+lr)+(j+ub)*ImageWidth];
							sum+=data16[pos-ImageWidth+1];

							//lr = 1;
							//ub = 0;
							//sum+=data[(i+lr)+(j+ub)*ImageWidth];
							sum+=data16[pos+1];

							//lr = 1;
							//ub = 1;
							//sum+=data[(i+lr)+(j+ub)*ImageWidth];
							sum+=data16[pos+ImageWidth+1];
				
							//sum+=data[(i+lr)+(j+ub)*ImageWidth];
				
							sum >>= 3;
				
							data16[pos]= (UINT16) sum;
						}// equalize
					} //if (datapos)
					pos++;
				} // for
			} // for2
		}
		else if (vinout->bpp == 32) // bpp == 32
		{
			UINT32 *data32 = (UINT32 *) vinout->framebits[curimg];
			INT32 j,i;
			
			for(j=up+1;j<=down-1;j++)
			{
				pos = j*ImageWidth+left+1;
	 
				for(i=left+1;i<=right-1;i++)
				{
					DOUBLE equalize=0;
					const UINT32 datapos = data32[pos];

					if (datapos)
					{
						const UINT32 datapos2 = (datapos / 2) * 3;
						//lr = -1;
						//ub = -1;
						//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
						equalize+=(DOUBLE) ((data32[pos-ImageWidth-1])<datapos2);

						//lr = -1;
						//ub = 0;
						//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
						equalize+=(DOUBLE) ((data32[pos-1])<datapos2);

						//lr = -1;
						//ub = 1;
						//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
						equalize+=(DOUBLE) ((data32[pos+ImageWidth-1])<datapos2);

						//lr=0;
						//ub = -1;
						//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
						equalize+=(DOUBLE) ((data32[pos-ImageWidth])<datapos2);

						//lr = 0;
						//ub = 1;
						//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
						equalize+=(DOUBLE) ((data32[pos+ImageWidth])<datapos2);

						//lr = 1;
						//ub = -1;
						//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
						equalize+=(DOUBLE) ((data32[pos-ImageWidth+1])<datapos2);

						//lr = 1;
						//ub = 0;
						//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
						equalize+=(DOUBLE) ((data32[pos+1])<datapos2);

						//lr = 1;
						//ub = 1;
						//equalize+=(data[(i+lr)+(j+ub)*ImageWidth]*2<data[i+ImageWidth*j]);
						equalize+=(DOUBLE) ((data32[pos+ImageWidth+1])<datapos2);

						if (equalize>3.0) //equalize
						{
							DOUBLE sum=0.0;

  							//lr = -1;
							//ub = -1;
							//sum+=data[(i+lr)+(j+ub)*ImageWidth];
							sum+=(DOUBLE) data32[pos-ImageWidth-1];

						    //lr = -1;
						    //ub = 0;
							//sum+=data[(i+lr)+(j+ub)*ImageWidth];
							sum+=(DOUBLE) data32[pos-1];

							//lr = -1;
							//ub = 1;
							//sum+=data[(i+lr)+(j+ub)*ImageWidth];
							sum+=(DOUBLE) data32[pos+ImageWidth-1];

							//lr=0;
							//ub = -1;
							//sum+=data[(i+lr)+(j+ub)*ImageWidth];
							sum+=(DOUBLE) data32[pos-ImageWidth];

							//lr = 0;
							//ub = 1;
							//sum+=data[(i+lr)+(j+ub)*ImageWidth];
							sum+=(DOUBLE) data32[pos+ImageWidth];

							//lr = 1;
							//ub = -1;
							//sum+=data[(i+lr)+(j+ub)*ImageWidth];
							sum+=(DOUBLE) data32[pos-ImageWidth+1];

							//lr = 1;
							//ub = 0;
							//sum+=data[(i+lr)+(j+ub)*ImageWidth];
							sum+=(DOUBLE) data32[pos+1];

							//lr = 1;
							//ub = 1;
							//sum+=data[(i+lr)+(j+ub)*ImageWidth];
							sum+=(DOUBLE) data32[pos+ImageWidth+1];
				
							//sum+=data[(i+lr)+(j+ub)*ImageWidth];
				
							sum /= 8.0;
				
							data32[pos]= (UINT32) sum;
						} // equalize
					} // if (datapos)
					pos++;
				} // for
	 		}//for2
		}
		else
		  return -1; // no proper bpp entry

    }// for all frames

	return 0;
}


// return value(s)
//	0 - okay, everything fine
// -1 - bad source data in vinout, data in vinout was NOT changed

VS_EXPORT INT32 normalizeframes_rect( struct VIDINFHDR *vinout, UINT32 x1, UINT32 y1, UINT32 x2, UINT32 y2 )
{
	UINT32 tmp;
	INT32 curimg;
	UINT32 effective_colorvalues = (1 << vinout->bpp_effective);


	if (vinout == NULL) return -1;
	if (vinout->type == TYPE_NONE_EMPTY) return -1;
		
	
	if (x1 > x2)
	{
		tmp = x1;
		x1 = x2;
		x2 = tmp;
	}

	if (y1 > y2)
	{
		tmp = y1;
		y1 = y2;
		y2 = tmp;
	}

	{

		UINT32 xadd = (vinout->width-1) - (x2-x1);
		register UINT32 xstart = y1*(vinout->width)+x1;
	
		register UINT32 x,y;
		UINT32 xpixels = x2 - x1 + 1;
		UINT32 ypixels = y2 - y1 + 1;
	
		for (curimg=0;curimg<vinout->numframes;curimg++)
		{
			if (vinout->bpp == 8)
			{
				register UCHAR *data = vinout->framebits[curimg];
				register UCHAR maxnum=0;
			
				for (y=0;y<ypixels;y++)
				{
					for (x=0;x<xpixels;x++)
					{
						if (data[xstart] > maxnum) maxnum = data[xstart];
						xstart++;
					}
					xstart+=xadd;
				}

				if (maxnum==0) maxnum=1;

				{
					register DOUBLE mulval = ((DOUBLE)(effective_colorvalues-1)) / (1.0*maxnum);

					xstart = y1*(vinout->width)+x1;

					for (y=0;y<ypixels;y++)
					{
						for (x=0;x<xpixels;x++)
						{
							data[xstart] = (UCHAR) (data[xstart]*mulval);
							xstart++;
						}
				
						xstart+=xadd;
					}
				}

			}
			else if (vinout->bpp == 16)
			{
				register UINT16 *data = vinout->framebits[curimg];
				register UINT16 maxnum=0;

				for (y=0;y<ypixels;y++)
				{
					for (x=0;x<xpixels;x++)
					{
						if (data[xstart] > maxnum) maxnum = data[xstart];
						xstart++;
					}
					xstart+=xadd;
				}

				if (maxnum==0) maxnum=1;

				{
					register DOUBLE mulval = ((DOUBLE)(effective_colorvalues-1)) / (1.0*maxnum);

					xstart = y1*(vinout->width)+x1;

					for (y=0;y<ypixels;y++)
					{
						for (x=0;x<xpixels;x++)
						{
							data[xstart] = (UINT16) (data[xstart]*mulval);
							xstart++;
						}
				
						xstart+=xadd;
					}
				}


			}
			else if (vinout->bpp == 32)
			{
				register UINT32 *data = vinout->framebits[curimg];
				register UINT32 maxnum=0;

				for (y=0;y<ypixels;y++)
				{
					for (x=0;x<xpixels;x++)
					{
						if (data[xstart] > maxnum) maxnum = data[xstart];
						xstart++;
					}
					xstart+=xadd;
				}

				if (maxnum==0) maxnum=1;

				{
					register DOUBLE mulval = ((DOUBLE)(effective_colorvalues-1)) / (1.0*maxnum);

					xstart = y1*(vinout->width)+x1;

					for (y=0;y<ypixels;y++)
					{
						for (x=0;x<xpixels;x++)
						{
							data[xstart] = (UINT32) (data[xstart]*mulval);
							xstart++;
						}
				
						xstart+=xadd;
					}
				}

			}
			else 
				return -1;

		}

	}
	return 0;

}






// return value(s)
//	0 - okay, everything fine
// -1 - bad source data in vinout, data in vinout was NOT changed
VS_EXPORT INT32 normalizeframes( struct VIDINFHDR *vinout)
{

	if (vinout == NULL) return -1;

	return normalizeframes_rect( vinout, 0, 0, vinout->width-1, vinout->height-1);

}
