
#include "platform.h"
#include "videofileload.h"
#include "significancemap.h"
#include <math.h>

// out must be passed uninitialized !
// sigma -> factor in calculation
//
// return codes:
//		 0		success
//		-1		out was not passed uninitialized
//		-2		signal_and_background must at least contain 2 frames
//		-3		background must at least contain 2 frames
//		-4		signal_and_bck, background must contain the same number of images!
//		-5		bitdepth, effective bitdepth, width or heigt of background and signal_and_background were
//				not set to equal values (important)
//		-6		memory allocation error, no data was changed (in out)	
//		-7		data error (unsupported bpp for algorithm)
//
INT32 calculate_significance_map( const struct VIDINFHDR *signal_and_background, const struct VIDINFHDR *background, struct VIDINFHDR *out, double sigma)
{
	if (out->type != TYPE_NONE_EMPTY) return (-1);

	if (signal_and_background->numframes < 2) return -2;
	if (background->numframes < 2) return -3;

	if (signal_and_background->numframes != background->numframes) return -4;


	if ( (signal_and_background->width != background->width) ||
		 (signal_and_background->height != background->height) ||
		 (signal_and_background->bpp != background->bpp) ||
		 (signal_and_background->bpp_effective != background->bpp_effective) ) return -5;

	if ((signal_and_background->bpp < 1) || (signal_and_background->bpp >32)) return -7;

	// now we can start

	{
		UINT32 numframes = signal_and_background->numframes;
		UINT32 bpp = signal_and_background->bpp;
		UINT32 i,j;
		UINT32 size_pixels = signal_and_background->width*signal_and_background->height;

		DOUBLE *sig_xmeanbuffer;
		DOUBLE *back_xmeanbuffer;
		DOUBLE  numimg_minus_1 = (numframes-1)*1.0;
		DOUBLE *rms_noise_buf;
		VOID   *resulting_buf;

		resulting_buf = (VOID *) malloc( size_pixels*bpp/8 );
		if (resulting_buf == NULL)
		{
			return -6;
		}

		memset(resulting_buf, 0, size_pixels*bpp/8);

		sig_xmeanbuffer = (DOUBLE *) malloc( size_pixels*sizeof(DOUBLE) );
		if (sig_xmeanbuffer == NULL) 
		{
			free( (void *) resulting_buf );
			return -6;
		}

		back_xmeanbuffer = (DOUBLE *) malloc( size_pixels*sizeof(DOUBLE) );
		if (back_xmeanbuffer == NULL) 
		{
			free( (void *) sig_xmeanbuffer );
			free( (void *) resulting_buf );
		
			return -6;
		}

		rms_noise_buf = (DOUBLE *) malloc( size_pixels*sizeof(DOUBLE) );
		if (rms_noise_buf == NULL) 
		{
			free( (void *) back_xmeanbuffer );
			free( (void *) sig_xmeanbuffer );
			free( (void *) resulting_buf );				

			return -6;
		}

		if (bpp <= 8)
		{
			// calculate xmean of signal+noise and of noise
			UCHAR *sig_bitmap = signal_and_background->framebits[0];
			UCHAR *back_bitmap = background->framebits[0];

			for (j=0;j<size_pixels;j++)
			{
				sig_xmeanbuffer[j] = (DOUBLE) (sig_bitmap[j]);
			}


			for (i=1;i<numframes;i++)
			{
				sig_bitmap = signal_and_background->framebits[i];
				
				for (j=0;j<size_pixels;j++)
				{
					sig_xmeanbuffer[j] += (DOUBLE) (sig_bitmap[j]);
				}
			}

			for (j=0;j<size_pixels;j++)
			{
				sig_xmeanbuffer[j] /= (DOUBLE) (numframes);
			}


			for (j=0;j<size_pixels;j++)
			{
				back_xmeanbuffer[j] = (DOUBLE) (back_bitmap[j]);
			}

			for (i=1;i<numframes;i++)
			{
				back_bitmap = background->framebits[i];
				
				for (j=0;j<size_pixels;j++)
				{
					back_xmeanbuffer[j] += (DOUBLE) (back_bitmap[j]);
				}
			}

			for (j=0;j<size_pixels;j++)
			{
				back_xmeanbuffer[j] /= (DOUBLE) (numframes);
			}

			// calculate RMS of noise

			for (j=0;j<size_pixels;j++)
			{
				DOUBLE bck_xmean = back_xmeanbuffer[j];
				DOUBLE result;

				result = 0.0;
				for (i=0;i<numframes;i++)
				{
					UCHAR *back_bitmap = background->framebits[i];
					DOUBLE result1 = bck_xmean - ((DOUBLE) back_bitmap[j]);

					result += ((result1*result1));
				}

				rms_noise_buf[j] = sigma*sqrt(result/numimg_minus_1);
			}

			{
				UCHAR *lresbuf = (UCHAR *) resulting_buf;

				// create resulting data set 
				for (j=0;j<size_pixels;j++)
				{
					if ((sig_xmeanbuffer[j]-back_xmeanbuffer[j])<=(rms_noise_buf[j])) lresbuf[j] = 1;
				}
			}

		}
		else if (bpp <= 16)
		{
			// calculate xmean of signal+noise and of noise
			UINT16 *sig_bitmap = signal_and_background->framebits[0];
			UINT16 *back_bitmap = background->framebits[0];

			for (j=0;j<size_pixels;j++)
			{
				sig_xmeanbuffer[j] = (DOUBLE) (sig_bitmap[j]);
			}


			for (i=1;i<numframes;i++)
			{
				sig_bitmap = signal_and_background->framebits[i];
				
				for (j=0;j<size_pixels;j++)
				{
					sig_xmeanbuffer[j] += (DOUBLE) (sig_bitmap[j]);
				}
			}

			for (j=0;j<size_pixels;j++)
			{
				sig_xmeanbuffer[j] /= (DOUBLE) (numframes);
			}


			for (j=0;j<size_pixels;j++)
			{
				back_xmeanbuffer[j] = (DOUBLE) (back_bitmap[j]);
			}

			for (i=1;i<numframes;i++)
			{
				back_bitmap = background->framebits[i];
				
				for (j=0;j<size_pixels;j++)
				{
					back_xmeanbuffer[j] += (DOUBLE) (back_bitmap[j]);
				}
			}

			for (j=0;j<size_pixels;j++)
			{
				back_xmeanbuffer[j] /= (DOUBLE) (numframes);
			}

			// calculate RMS of noise

			for (j=0;j<size_pixels;j++)
			{
				DOUBLE bck_xmean = back_xmeanbuffer[j];
				DOUBLE result;

				result = 0.0;
				for (i=0;i<numframes;i++)
				{
					UINT16 *back_bitmap = background->framebits[i];
					DOUBLE result1 = bck_xmean - ((DOUBLE) back_bitmap[j]);

					result += ((result1*result1));
				}

				rms_noise_buf[j] = sigma*sqrt(result/numimg_minus_1);
			}

			{
				UINT16 *lresbuf = (UINT16 *) resulting_buf;

				// create resulting data set 
				for (j=0;j<size_pixels;j++)
				{
					if ((sig_xmeanbuffer[j]-back_xmeanbuffer[j])<=(rms_noise_buf[j])) lresbuf[j] = 1;
				}
			}

		}
		else if (bpp <= 32)
		{
			// calculate xmean of signal+noise and of noise
			UINT32 *sig_bitmap = signal_and_background->framebits[0];
			UINT32 *back_bitmap = background->framebits[0];

			for (j=0;j<size_pixels;j++)
			{
				sig_xmeanbuffer[j] = (DOUBLE) (sig_bitmap[j]);
			}


			for (i=1;i<numframes;i++)
			{
				sig_bitmap = signal_and_background->framebits[i];
				
				for (j=0;j<size_pixels;j++)
				{
					sig_xmeanbuffer[j] += (DOUBLE) (sig_bitmap[j]);
				}
			}

			for (j=0;j<size_pixels;j++)
			{
				sig_xmeanbuffer[j] /= (DOUBLE) (numframes);
			}


			for (j=0;j<size_pixels;j++)
			{
				back_xmeanbuffer[j] = (DOUBLE) (back_bitmap[j]);
			}

			for (i=1;i<numframes;i++)
			{
				back_bitmap = background->framebits[i];
				
				for (j=0;j<size_pixels;j++)
				{
					back_xmeanbuffer[j] += (DOUBLE) (back_bitmap[j]);
				}
			}

			for (j=0;j<size_pixels;j++)
			{
				back_xmeanbuffer[j] /= (DOUBLE) (numframes);
			}

			// calculate RMS of noise

			for (j=0;j<size_pixels;j++)
			{
				DOUBLE bck_xmean = back_xmeanbuffer[j];
				DOUBLE result;

				result = 0.0;
				for (i=0;i<numframes;i++)
				{
					UINT32 *back_bitmap = background->framebits[i];
					DOUBLE result1 = bck_xmean - ((DOUBLE) back_bitmap[j]);

					result += ((result1*result1));
				}

				rms_noise_buf[j] = sigma*sqrt(result/numimg_minus_1);
			}

			{
				UINT32 *lresbuf = (UINT32 *) resulting_buf;

				// create resulting data set 
				for (j=0;j<size_pixels;j++)
				{
					if ((sig_xmeanbuffer[j]-back_xmeanbuffer[j])<=(rms_noise_buf[j])) lresbuf[j] = 1;
				}
			}
		}
		else return -7; // no proper bpp setting

		out->framescales = (DOUBLE *) malloc( 1*sizeof(DOUBLE) );
		if (out->framescales == NULL)
		{
			free( (void *) rms_noise_buf );
			free( (void *) back_xmeanbuffer );
			free( (void *) sig_xmeanbuffer );
			free( resulting_buf );

			return -6;
		}

		out->framebits = (VOID **) malloc( sizeof( VOID **)*1 );
		if (out->framebits == NULL)
		{
			free( (void *) out->framescales );
			free( (void *) rms_noise_buf );
			free( (void *) back_xmeanbuffer );
			free( (void *) sig_xmeanbuffer );
			free( resulting_buf );

			return -6;
		}

		out->framebits[0] = (void *) resulting_buf;
		out->framescales[0] = signal_and_background->scale;

		out->bpp = signal_and_background->bpp;
		out->bpp_effective = signal_and_background->bpp_effective;
		out->scale = signal_and_background->scale;
		out->type = TYPE_SIGNIFICANCE;
		out->width = signal_and_background->width;
		out->height = signal_and_background->height;
		out->numframes = 1;

		free( (void *) rms_noise_buf );
		free( (void *) back_xmeanbuffer );
		free( (void *) sig_xmeanbuffer );

	}


	return 0;
}
