////////////////////////////////////////////////////////////////////////////////////
//
// videofileload.cpp  
//
//  functions to load in certain types of files created by the video system at PITZ
//
//		IMM		-	images (single or multiple), uncompressed
//		IMC		-	images (single or multiple), compressed (zlib compression)
//		BKG		-	background image (single), uncompressed
//		BKC		-	raw background images (single or multiple), compressed (zlib)
//
//	A save function is provided to save a loaded in file to IMM format. (as an
//	example)
//
////////////////////////////////////////////////////////////////////////////////////
//	added save functionality for imm, bkg, imc and bkc, Stefan Weisse (July 2006)
////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////
//	created and modified by Stefan Weisse / DESY Zeuthen, November 2005
////////////////////////////////////////////////////////////////////////////////////


#include <stdio.h>
#include "platform.h"
#include "zlibcompr_excerpt.h"
#include "videofileload.h"

////////////////////////////////////////////////////////////////////////////////////
// wrapper function for erasing file on disk
////////////////////////////////////////////////////////////////////////////////////
//  -1	 error condition
//   0	 file was successfully deleted.
//
INT32 erase_file_disk( FILE *fp, const CHAR *filename)
{
	
	if (fp != NULL) fclose( fp );
	

	return( remove( filename ) );
}


////////////////////////////////////////////////////////////////////////////////////
//
// String handling for platforms where no _strnicmp (string compare case insensitive 
// version of strncmp) is provided
//
////////////////////////////////////////////////////////////////////////////////////

#define MAXLEN 1024

////////////////////////////////////////////////////////////////////////////////////
//	convert a given string (*s, array of chars, \0) to upper case 
////////////////////////////////////////////////////////////////////////////////////
void stoupper(char *s)
{
    char *p = s;
    
    if (p == NULL) return;
    
    while (*p != 0)
    {
    	if( islower(*p) ) *p = _toupper(*p);
	++p;
    }
}



char strtmp1[MAXLEN];
char strtmp2[MAXLEN];

////////////////////////////////////////////////////////////////////////////////////
//	implementation of _strnicmp using stoupper
////////////////////////////////////////////////////////////////////////////////////
int _strnicmp( const char *string1, const char *string2, size_t count)
{
 if (count > (MAXLEN-1)) return -1;
 
 strncpy(strtmp1, string1, count);
 strtmp1[count] = 0;
 
 strncpy(strtmp2, string2, count);
 strtmp2[count] = 0;
 
 stoupper(strtmp1); 
 stoupper(strtmp2);
 
 return strncmp( strtmp1, strtmp2, count );
} 



////////////////////////////////////////////////////////////////////////////////////
//
//	Initializes a newly defined (instantiated) VIDINFHDR
//
//  This function never fails.
//
////////////////////////////////////////////////////////////////////////////////////
VS_EXPORT void initialize_vidinfhdr( struct VIDINFHDR *vinf )
{
    vinf->type = TYPE_NONE_EMPTY;
	vinf->framebits = NULL;
	vinf->framescales = NULL;
}


////////////////////////////////////////////////////////////////////////////////////
//
// Function to free a VIDINFHDR which might contain dynamic memory attached.
//
// This function never fails.
//
////////////////////////////////////////////////////////////////////////////////////
VS_EXPORT void free_vidinfhdr( struct VIDINFHDR *vinf )
{
	INT32 i;
	
	if ((vinf->type == TYPE_IMM) || (vinf->type == TYPE_IMC) || (vinf->type == TYPE_BKG) || (vinf->type == TYPE_BKC) || (vinf->type == TYPE_SIGNIFICANCE))
	{
		vinf->type = TYPE_NONE_EMPTY;
		
		if (vinf->framescales != NULL) free( (void *) vinf->framescales );

		if (vinf->framebits != NULL)
		{

			for (i=0;i<vinf->numframes;i++)
			{
				if (vinf->framebits[i] != NULL) free( (void *) vinf->framebits[i] );
			}

			free( (void *) vinf->framebits );
		}
		
		vinf->framebits = NULL;
		vinf->framescales = NULL;
	}
}


////////////////////////////////////////////////////////////////////////////////////
//
// This function loads in an imm file (pointed to by filename) into the 
// structure vinf (VIDINFHDR).
//
// The return values are:
//
//	 0  successful
//
//	-3	could not open file
//	-4	error in file !
//	-5	file read error !
//	-6	already initialized, please free VIDINFHDR (and memory) first
//	-7	out of memory
//
////////////////////////////////////////////////////////////////////////////////////
VS_EXPORT INT32 load_imm_file( CHAR *filename, struct VIDINFHDR *vinf )
{
	FILE *fp;
	UINT32 w, h, bpp, bpp_eff;
	UINT32 ret;
	INT32 num_images;

	if (vinf->framescales != NULL) return( -6 );
	if (vinf->framebits != NULL) return( -6 );

	fp = fopen(filename, "rb");

	if (fp == NULL) return( -3 );

	ret = fread( &w, 1, 4, fp );
	ret += fread( &h, 1, 4, fp );
   
	// mask out bpp, store as bits per pixel
	bpp = (w&0x00FF0000) >> 16;
	bpp_eff = (h&0x00FF0000) >> 16;
	// convert back to real width
	w = w&0xffff;
	h = h&0xffff;

	// old file: bpp == 0 so make it 8
	if (bpp == 0) bpp = 8;
	if (bpp_eff == 0) bpp_eff = 8;

	if (ret == 8)
	{
		INT32 size;
		INT32 i;

		fseek( fp, 0, SEEK_END);
		size = ftell(fp);
		num_images = size / (((w*h*bpp)/8)+8+8 );
		fseek( fp, 0, SEEK_SET);

		if (num_images == 0) { fclose(fp); return( -4 ); }

		vinf->framescales = (DOUBLE *) malloc( num_images*sizeof(DOUBLE) );
		vinf->framebits = (VOID **) malloc( sizeof( VOID **)*num_images );

		if ((vinf->framescales) == NULL || (vinf->framebits == NULL))
		{
			fclose(fp);
			return( -7 );
		}
 
		for (i=0;i<num_images;i++)
		{	
			UINT32 lw, lh, lbpp, lbpp_eff;
			UINT32 ret;
     
			// read in width and height	 
			ret = fread(&lw, 1, 4, fp);
			ret += fread(&lh, 1, 4, fp);

			// mask out bpp, store as bits per pixel
			lbpp = (lw&0x00FF0000) >> 16;
			lbpp_eff = (lh&0x00ff0000) >> 16;

			// convert back to real width
			lw = lw&0xffff;
			lh = lh&0xffff;
			
			if (lbpp == 0) lbpp = 8;
			if (lbpp_eff == 0) lbpp_eff = 8;
       
			if ((lw != w) || (lh != h) || (lbpp != bpp) || (lbpp_eff != bpp_eff)) // check for right width and height and bpp and bpp_effective
			{
				fclose(fp);
				return( -4 );
			}

			vinf->framebits[i] = (void *) malloc( lw*lh*lbpp/8 );

			if (vinf->framebits[i] == NULL) { fclose(fp); return( -7 ); }
		
			ret += fread( vinf->framebits[i], 1, lw*lh*lbpp/8, fp);
			ret += fread( &(vinf->framescales[i]), 1, sizeof(DOUBLE), fp);

			if (ret != (16+(lw*lh*lbpp/8)) ) { fclose(fp); return( -5 ); }
		}

		vinf->bpp = bpp;
		vinf->bpp_effective = bpp_eff;
		vinf->numframes = num_images;
		vinf->height = h;
		vinf->width = w;
		vinf->scale = vinf->framescales[0];
		vinf->type = TYPE_IMM;

 	}
	else 
	{ 
		fclose(fp); 
		return( -5 ); 
	}

	fclose(fp);

	return 0;
}

////////////////////////////////////////////////////////////////////////////////////
//
// This function loads in an bkg file (pointed to by filename) into the 
// structure vinf (VIDINFHDR).
//
// The return values are:
//
//	 0	successful
//
//	-3	could not open file
//	-5	file read error !
//	-6	already initialized, please free VIDINFHDR (and memory) first
//	-7	out of memory
//
////////////////////////////////////////////////////////////////////////////////////
VS_EXPORT INT32 load_bkg_file( CHAR *filename, struct VIDINFHDR *vinf )
{
	FILE *fp;
	UINT32 w, h, bpp, bpp_eff;
	UINT32 ret;

	if (vinf->framescales != NULL) return( -6 );
	if (vinf->framebits != NULL) return( -6 );

	fp = fopen(filename, "rb");

	if (fp == NULL) return( -3 );

	ret = fread( &w, 1, 4, fp );
	ret += fread( &h, 1, 4, fp );
   
	// mask out bpp, store as bits per pixel
	bpp = (w&0x00FF0000) >> 16;
	bpp_eff = (h&0x00FF0000) >> 16;

	// convert back to real width
	w = w&0xffff;
	h = h&0xffff;

	// old file: bpp == 0 so make it 8
	if (bpp == 0) bpp = 8;
	if (bpp_eff == 0) bpp_eff = 8;

	if (ret == 8)
	{
		UINT32 len = (w*h*bpp) / 8;

		vinf->framescales = (DOUBLE *) malloc( 1*sizeof(DOUBLE) );
		vinf->framebits = (VOID **) malloc( sizeof( VOID **)*1 );

		if ((vinf->framescales) == NULL || (vinf->framebits == NULL))
		{
			fclose(fp);
			return( -7 );
		}
 
		vinf->framebits[0] = (VOID *) malloc( len );

		if (vinf->framebits[0] == NULL) { fclose(fp); return( -7 ); }

		vinf->framescales[0] = 1.0000;
		
		ret += fread( vinf->framebits[0], 1, len, fp);

		if (ret != (8+len) ) { fclose(fp); return( -5 ); }

		vinf->bpp = bpp;
		vinf->bpp_effective = bpp_eff;
		vinf->numframes = 1;
		vinf->height = h;
		vinf->width = w;
		vinf->scale = vinf->framescales[0];
		vinf->type = TYPE_BKG;

 	}
	else 
	{ 
		fclose(fp); 
		return( -5 ); 
	}

	fclose(fp);

	return 0;
}


////////////////////////////////////////////////////////////////////////////////////
//
// This function loads in either an imc or bkc file (structures inside the file
// are the same, only the extension is differnet to distinguish between backgrounds 
// and images) into the structure vinf (VIDINFHDR).
//
// The return values are:
//
//	 0	successful
//
//	-3	could not open file
//	-4	error in file ! (e.g. file corrupted)
//	-5	file read error !
//	-6	already initialized, please free VIDINFHDR (and memory) first
//	-7	out of memory
//	-8	decompression error
//
////////////////////////////////////////////////////////////////////////////////////
VS_EXPORT INT32 load_imc_bkc_file( CHAR *filename, struct VIDINFHDR *vinf, INT32 type)
{
	FHDR fHdr;
	IMHDR imHdr;
	INT32 ret;
	FILE *fp;
	UINT32 i;

	if (vinf->framescales != NULL) return( -6 );
	if (vinf->framebits != NULL) return( -6 );

	fp = fopen(filename, "rb");

	if (fp == NULL) return( -3 );

	// read in file header
	ret = fread( &fHdr, 1, sizeof(FHDR), fp);

	if (ret != sizeof(FHDR))
	{
		fclose(fp);
		return( -5 );
	}

	// upper 16 bits of file header 'bpp' filled -> this is the effective bpp
	if (fHdr.bpp&0xffff0000)
	{
		vinf->bpp_effective = (fHdr.bpp>>16)&0xffff;
		vinf->bpp = fHdr.bpp & 0xffff;
	}
	else
	{
		vinf->bpp = fHdr.bpp;
		vinf->bpp_effective = vinf->bpp;
	}

	if ((vinf->bpp != 8) && (vinf->bpp != 16) && (vinf->bpp != 32))
	{
		fclose(fp);
		return( -4 );
	}

	vinf->framescales = (DOUBLE *) malloc( fHdr.num_frames*sizeof(DOUBLE) );
	vinf->framebits = (VOID **) malloc( sizeof( VOID *)*fHdr.num_frames );

	if ((vinf->framescales) == NULL || (vinf->framebits == NULL))
	{
		fclose(fp);
		return( -7 );
	}

	vinf->width = fHdr.width;
	vinf->height = fHdr.height;
	vinf->numframes = fHdr.num_frames;

	// loop over number of frames
	for (i=0;i<fHdr.num_frames;i++)
	{
		UCHAR *compressed;
		UCHAR *uncompr_data;
		uLong len_uncompressed;
		INT32 retc;

		// read in image header of image # i 
		ret = fread( &imHdr, 1, sizeof(IMHDR), fp);

		if (ret != sizeof(IMHDR)) // error while reading ? -> clean up and close
		{
			fclose(fp);
			return( -5 );
		}

		compressed = (UCHAR *) malloc( imHdr.len_compressed );
		if (compressed == NULL) // error while allocating ? -> clean up and close
		{
		
			fclose(fp);
			return -7;
		}

		ret = fread( (VOID *) compressed, 1, imHdr.len_compressed, fp);
		if (ret != ((INT32)imHdr.len_compressed)) // error while reading ? -> clean up and close
		{
			free( (VOID *) compressed );
			fclose(fp);
			return( -7 );
		}

		vinf->framescales[i] = imHdr.scale;
		

		vinf->framebits[i] = (VOID *) malloc( imHdr.len_uncompressed );
		if (vinf->framebits[i] == NULL) // error while allocating ? -> clean up and close
		{
   			free( (VOID *) compressed);
			fclose(fp);
			return( -7 );
		}
 
		uncompr_data = (UCHAR *) vinf->framebits[i];

		len_uncompressed = (uLong) imHdr.len_uncompressed;

		// uncompress video frame bits
		retc = uncompress ((Bytef *) uncompr_data, &len_uncompressed, (const Bytef *) compressed, (uLong) imHdr.len_compressed);
		if (retc != Z_OK)
		{
		
			free( (VOID *) vinf->framebits[i] );
			free( (VOID *) compressed );
			fclose(fp);
			return( -8 );
		}

		free( (VOID *) compressed );
	}

  	vinf->scale = vinf->framescales[0];
	vinf->type = type;

	return 0;
}


////////////////////////////////////////////////////////////////////////////////////
//
// This function loads in an imc file (given by filename) into the structure 
// vinf (VIDINFHDR).
//
// The return values are:
//
//	 0	successful
//
//	-3	could not open file
//	-4	error in file ! (e.g. file corrupted)
//	-5	file read error !
//	-6	already initialized, please free VIDINFHDR (and memory) first
//	-7	out of memory
//	-8	decompression error
//
////////////////////////////////////////////////////////////////////////////////////
VS_EXPORT INT32 load_imc_file( CHAR *filename, struct VIDINFHDR *vinf )
{
	return load_imc_bkc_file( filename, vinf, TYPE_IMC );
}


////////////////////////////////////////////////////////////////////////////////////
//
// This function loads in a bkc file (given by filename) into the structure 
// vinf (VIDINFHDR).
//
// The return values are:
//
//	 0	successful
//
//	-3	could not open file
//	-4	error in file ! (e.g. file corrupted)
//	-5	file read error !
//	-6	already initialized, please free VIDINFHDR (and memory) first
//	-7	out of memory
//	-8	decompression error
//
////////////////////////////////////////////////////////////////////////////////////
VS_EXPORT INT32 load_bkc_file( CHAR *filename, struct VIDINFHDR *vinf )
{
	return load_imc_bkc_file( filename, vinf, TYPE_BKC );
}


////////////////////////////////////////////////////////////////////////////////////
//
// This function loads in an image file (given by filename) created by the video 
// system into the structure vinf (VIDINFHDR). 
//
// One may either choose the type by itself (specify the right type, e.g. TYPE_IMM) 
// or the function tries to guess the file type by the extension of the filename
// and calls the proper file loading function afterwards (use TYPE_CHOOSE).
//
//	Parameters:
//		type	-	gives file type, valid types are 
//						TYPE_IMM		- specifies an IMM file
//						TYPE_IMC		- specifies an IMC file
//						TYPE_BKG		- specifies an BKG file
//						TYPE_BKC		- specifies an BKC file
//						TYPE_CHOOSE		- let the function try to find out the type of
//										  the file by itself
//
//	  filename	-	specifies the location of the file
//
//		vinf	-	VIDINFHDR structure where to put the loaded in file data
//
// The return values are:
//
//	 0	successful
//
//	-1	file could not be checked for type (by extension)
//	-2	file type unknown
//	-3	could not open file
//	-4	error in file ! (e.g. file corrupted)
//	-5	file read error !
//	-6	already initialized, please free VIDINFHDR (and memory) first
//	-7	out of memory
//	-8	decompression error
//
////////////////////////////////////////////////////////////////////////////////////
VS_EXPORT INT32 load_file( INT32 type, CHAR *filename, struct VIDINFHDR *vinf )
{
	if (type == TYPE_CHOOSE)
	{
		CHAR *pos = strrchr(filename, '.' );

		if (pos == NULL) return( -1 );

		pos++;

		if (*pos == 0) return( -1 );

		if (_strnicmp(pos, "IMM", 3) == 0) type = TYPE_IMM;
		else if (_strnicmp(pos, "IMC", 3) == 0) type = TYPE_IMC;
		else if (_strnicmp(pos, "BKG", 3) == 0) type = TYPE_BKG;
		else if (_strnicmp(pos, "BKC", 3) == 0) type = TYPE_BKC;
		else return( -1 );

	}

	switch( type )
	{
		case TYPE_IMM:
			return ( load_imm_file( filename, vinf ) );

			break;


		case TYPE_IMC:
			return ( load_imc_file( filename, vinf ) );

			break;

		
		case TYPE_BKG:
			return ( load_bkg_file( filename, vinf ) );

			break;


		case TYPE_BKC:
			return ( load_bkc_file (filename, vinf) );

			break;

		default: break; 
	}


	return( -2 );

}

////////////////////////////////////////////////////////////////////////////////////
//
// This function saves an image file (given by filename) out of the structure 
// vinf (VIDINFHDR). 
//
// One has to choose the type by itself (specify the right type, e.g. TYPE_IMM).
// The image data inside the structure vinf will be checked for validity, e.g.
// a BKG file can only contain a single frame.
//
//	Parameters:
//		type	-	gives file type, valid types are 
//						TYPE_IMM		- specifies an IMM file
//						TYPE_IMC		- specifies an IMC file
//						TYPE_BKG		- specifies an BKG file
//						TYPE_BKC		- specifies an BKC file
//						TYPE_CHOOSE		- choose type to save by content (type field) of vinf structure
//										  add extension to filename automatically based on type!
//
//	  filename	-	specifies the location of the file
//
//		vinf	-	VIDINFHDR structure where to put the loaded in file data
//
// The return values are:
//
//	 0	successful
//
//	-1	multiple video frames can not be saved as bkg file
//	-2	file type unknown
//	-3  error in vinf structure (either bad type or bad numframes (i.e. 0) )
	
//	-4	could not open file
//	-5	file write error !
//	-6	out of memory
//	-7	compression error
//
////////////////////////////////////////////////////////////////////////////////////
VS_EXPORT INT32 save_file( INT32 type, CHAR *filename, struct VIDINFHDR *vinf )
{
	char lFilename[MAXLEN];

	if (strlen(filename) > MAXLEN-5) return -8;

	if (vinf == NULL) return -3;
	if (vinf->numframes == 0) return -3;

	if (type == TYPE_CHOOSE)
	{
		type = vinf->type;
		memset(lFilename, 0, MAXLEN);
		strncpy(lFilename, filename, MAXLEN-5);
		lFilename[MAXLEN-5] = 0;

		if (type == TYPE_IMM) strcat(lFilename, ".imm");
		else if (type == TYPE_IMC) strcat( lFilename, ".imc");
		else if (type == TYPE_BKG) strcat( lFilename, ".bkg");
		else if (type == TYPE_BKC) strcat( lFilename, ".bkc");
		else return -2;
	}
	else
	{
		strncpy(lFilename, filename, strlen(filename)+1);
	}
	
	switch( type )
	{
		case TYPE_IMM:
			return ( save_imm_file( lFilename, vinf ) );

			break;


		case TYPE_IMC:
			return ( save_imc_file( lFilename, vinf ) );

			break;

		
		case TYPE_BKG:
			if (vinf->numframes > 1) return( -1 );
			
			return ( save_bkg_file( lFilename, vinf ) );

			break;


		case TYPE_BKC:
			return ( save_bkc_file (lFilename, vinf) );

			break;

		default: 
			break; 
	}


	return( -2 );

}

////////////////////////////////////////////////////////////////////////////////////
//
// This function writes the image data inside a passed vinf (VIDINFHDR) into a 
// newly created IMM file with name 'filename'. 
//
// The return values are:
//
//	 0	successful
//
//	-2	file type unknown
//	-3  error in vinf structure (bad numframes (i.e. 0), vinf == NULL, ... )
//	-4	could not open file for writing
//	-5	file write error !
//
////////////////////////////////////////////////////////////////////////////////////
VS_EXPORT INT32 save_imm_file( CHAR *filename, struct VIDINFHDR *vinf )
{
	FILE *fp;
	UINT32 w_wri,h_wri,frmsiz;
	DOUBLE sc;
	UINT32 ret;
	INT32 i;

	if (vinf == NULL) return -3;
	if ((vinf->type != TYPE_IMM) && (vinf->type != TYPE_IMC) && (vinf->type != TYPE_BKG) && (vinf->type != TYPE_BKC))
	{
		return -2;
	}
	if (vinf->numframes == 0) return( -3 );

	frmsiz = (vinf->width*vinf->height*vinf->bpp) / 8;

	// for safety
	if (vinf->bpp_effective == 0) vinf->bpp_effective = vinf->bpp;

	if (vinf->bpp == 8)
	{
		vinf->bpp_effective = 0;
		vinf->bpp = 0;
	}

	w_wri = ((vinf->bpp&0xFF) << 16) + (vinf->width&0xFFFF);
	h_wri = ((vinf->bpp_effective&0xFF) << 16) + (vinf->height&0xFFFF);
	sc = vinf->scale;
	
	fp = fopen( filename, "wb");
	if (fp == NULL) return( -4 );

	for (i=0;i<vinf->numframes;i++)
	{
		ret = fwrite( &w_wri, 1, sizeof(UINT32), fp);
		ret += fwrite( &h_wri, 1, sizeof(UINT32), fp);

		ret += fwrite( vinf->framebits[i], 1, frmsiz, fp);
		ret += fwrite( &vinf->framescales[i], 1, sizeof(DOUBLE), fp);

		if (ret != (sizeof(UINT32)*2+frmsiz+sizeof(DOUBLE)))
		{
			erase_file_disk( fp, filename );
			return(-5);
		}
	}
	fclose(fp);

	return 0;
}

////////////////////////////////////////////////////////////////////////////////////
//
// This function writes the image data inside a passed vinf (VIDINFHDR) into a 
// newly created BKG file with name 'filename'. 
//
// The return values are:
//
//	 0	successful
//
//	-1	multiple video frames can not be saved as bkg file
//	-2	file type unknown
//	-3  error in vinf structure (bad numframes (i.e. 0), vinf == NULL, ... )
//	-4	could not open file for writing
//	-5	file write error !
//
////////////////////////////////////////////////////////////////////////////////////
VS_EXPORT INT32 save_bkg_file( CHAR *filename, struct VIDINFHDR *vinf )
{
	FILE *fp;
	UINT32 w_wri,h_wri,frmsiz;
	UINT32 ret;

	if (vinf == NULL) return -3;
	if (vinf->numframes != 1) return( -1 );
	if ((vinf->type != TYPE_IMM) && (vinf->type != TYPE_IMC) && (vinf->type != TYPE_BKG) && (vinf->type != TYPE_BKC))
	{
		return -2;
	}

	frmsiz = (vinf->width*vinf->height*vinf->bpp) / 8;

	// for safety
	if (vinf->bpp_effective == 0) vinf->bpp_effective = vinf->bpp;

	if (vinf->bpp == 8)
	{
		vinf->bpp_effective = 0;
		vinf->bpp = 0;
	}

	w_wri = ((vinf->bpp&0xFF) << 16) + (vinf->width&0xFFFF);
	h_wri = ((vinf->bpp_effective&0xFF) << 16) + (vinf->height&0xFFFF);

	fp = fopen( filename, "wb");
	if (fp == NULL) return( -4 );

	ret = fwrite( &w_wri, 1, sizeof(UINT32), fp);
	ret += fwrite( &h_wri, 1, sizeof(UINT32), fp);

	ret += fwrite( vinf->framebits[0], 1, frmsiz, fp);

	if (ret != (sizeof(UINT32)*2+frmsiz))
	{
		erase_file_disk( fp, filename );
		return(-5);
	}

	fclose(fp);

	return 0;
}



////////////////////////////////////////////////////////////////////////////////////
//
// This function writes the image data inside a passed vinf (VIDINFHDR) into a 
// newly created BKC file with name 'filename'. 
//
// The return values are:
//
//	 0	successful
//
//	-2	file type unknown
//	-3  error in vinf structure (bad numframes (i.e. 0), vinf == NULL, ... )
//	-4	could not open file for writing
//	-5	file write error !
//	-6	out of memory
//	-7	compression error
//
////////////////////////////////////////////////////////////////////////////////////

VS_EXPORT INT32 save_bkc_file(CHAR *filename, struct VIDINFHDR *vinf)
{
	return save_imc_file(filename, vinf);
}


////////////////////////////////////////////////////////////////////////////////////
//
// This function writes the image data inside a passed vinf (VIDINFHDR) into a 
// newly created IMC file with name 'filename'. 
//
// The return values are:
//
//	 0	successful
//
//	-2	file type unknown
//	-3  error in vinf structure (bad numframes (i.e. 0), vinf == NULL, ... )
//	-4	could not open file for writing
//	-5	file write error !
//	-6	out of memory
//	-7	compression error
//
////////////////////////////////////////////////////////////////////////////////////
VS_EXPORT INT32 save_imc_file(CHAR *filename, struct VIDINFHDR *vinf)
{
	FILE *fp;
	FHDR fHdr;
	IMHDR imHdr;
	UINT32 ret;
	INT32 cnt;
	CHAR *compressed;
	uLong comprlen;

	if (vinf == NULL) return -3;
	if ((vinf->type != TYPE_IMM) && (vinf->type != TYPE_IMC) && (vinf->type != TYPE_BKG) && (vinf->type != TYPE_BKC))
	{
		return -2;
	}
	if (vinf->numframes == 0) return( -3 );

	fp = fopen(filename, "wb");
	if (fp == NULL) // could not create file: error
	{
		return -4;
	}

	if (vinf->bpp_effective == 0) vinf->bpp_effective = 8;
	if (vinf->bpp == 8) vinf->bpp_effective = 0;

	fHdr.width = vinf->width;
	fHdr.height = vinf->height;
	fHdr.bpp = ((vinf->bpp_effective<<16)&0xffff0000) + (vinf->bpp&0xffff);
	fHdr.num_frames = vinf->numframes;

	ret = fwrite( &fHdr, 1, sizeof(FHDR), fp);
	if (ret != sizeof(FHDR) ) // error while writing ? 
	{
		erase_file_disk( fp, filename ); // close and erase zombie
		return -5;
	}

	for (cnt = 0; cnt < vinf->numframes; cnt++ )
	{
		imHdr.scale = vinf->framescales[cnt];
		imHdr.len_uncompressed = (vinf->width*vinf->height*vinf->bpp) / 8;
		
		comprlen= ((uLong)imHdr.len_uncompressed*3) / 2;

		// allocate memory for temorarily storing the compressed frame bits
		compressed = (CHAR *) malloc( sizeof(CHAR)*comprlen );
		if (compressed == NULL) // error allocating ? 
		{
			erase_file_disk( fp, filename ); // close and erase zombie
			return -6; // -> return error
		}
  
		// compress the frame using zlib
		ret = compress ((Bytef *)compressed, (uLongf *) &comprlen, (const Bytef *) (vinf->framebits[cnt]), (uLong) imHdr.len_uncompressed); 
		if (ret != Z_OK) // error ? 
		{
			free( (void *) compressed );
			erase_file_disk( fp, filename ); // close and erase zombie
			return -7; // return error
		}

		// fill out last remaining field of image header
		imHdr.len_compressed = (INT32) comprlen;

		// write out image header
		ret = fwrite( &imHdr, 1, sizeof(IMHDR), fp);
		if (ret != sizeof(IMHDR) ) // error ? 
		{
			free( (void *) compressed );
			erase_file_disk( fp, filename ); // close and erase zombie
			return -5; // return error
		}

		// write out compressed frame bits
		ret = fwrite( compressed, 1, imHdr.len_compressed, fp); 
		if (ret != comprlen) // error ?  
		{
			free( (void *) compressed );
			erase_file_disk( fp, filename ); // close and erase zombie
			return -5; // return error
		}

		// free compressed ptr
		free( (void *) compressed );
	}

	fclose(fp);
	return 0;
}


